#include "../../std.h"

#include "../../misc/inttostring.h"
#include "../../misc/url_options.h"

#include "srm_url.h"

SRM_URL::SRM_URL(std::string url) try: URL(url) {
  portdefined = false;
  if(protocol != "srm") { valid=false; return; };
  valid=true;
  if(port <= 0) port=8443;
  else portdefined = true;
  srm_version = SRM_URL_VERSION_2_2; // v2.2 unless explicitly made otherwise
  std::string::size_type p = path.find("?SFN=");
  if(p != std::string::npos) {
    filename=path.c_str()+p+5;
    path.resize(p);
    isshort=false;
    for(;path.size() > 1;) {
      if(path[1] != '/') break;
      path.erase(0,1);
    };
    if (path[path.size()-1] == '1') srm_version = SRM_URL_VERSION_1;
  }
  else {
    if(path.length() > 0) filename=path.c_str()+1; // Skip leading '/'
    path = "/srm/managerv2";
    isshort=true;
  }
} catch (std::exception e) {
  valid=false;
}

std::string SRM_URL::empty("");

void SRM_URL::SetSRMVersion(std::string version) {
  if (version.empty()) return;
  if (version == "1") {
    srm_version = SRM_URL_VERSION_1;
    path = "/srm/managerv1";
  }
  else if (version == "2.2") {
    srm_version = SRM_URL_VERSION_2_2;
    path = "/srm/managerv2";
  }
  else {
    srm_version = SRM_URL_VERSION_UNKNOWN; 
  }
}

std::string SRM_URL::ContactURL(void) const {
  if(!valid) return empty;
  return ("httpg://"+host+":"+inttostring(port)+path);
}

std::string SRM_URL::BaseURL(void) const {
  if(!valid) return empty;
  return (protocol+"://"+host+":"+inttostring(port)+path+"?SFN=");
}

std::string SRM_URL::FullURL(void) const {
  if(!valid) return empty;
  return (protocol+"://"+host+":"+inttostring(port)+path+"?SFN="+filename);
}

std::string SRM_URL::ShortURL(void) const {
  return (protocol+"://"+host+":"+inttostring(port)+"/"+filename);
}

void SRM_URL::GSSAPI(bool gssapi) {
  if (gssapi)
    urloptions["protocol"] = "gssapi";
  else
    urloptions["protocol"] = "gsi";
}

bool SRM_URL::GSSAPI(void) const {
  try {
    const std::string proto_val =
          ((std::map<std::string,std::string>&)(Options()))["protocol"];
    if(proto_val == "gsi") return false;
  } catch (std::exception e) { };
  return true;
}
